#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <sys/time.h>

/******************************************************/
/**The code is used for measuring time of the main cpu*/
/******************************************************/

void fill_int(uint32_t *a, uint32_t length)
{
  uint32_t i;

  for(i=0; i<length; ++i) {  
    a[i] = i;
  }

}

void fill_fp(float *a, uint32_t length)
{
  uint32_t i;

  for(i=0; i<length; ++i) {  
    a[i] = (i+10.0)/(i+1.0);
  }

}


int dsblock_payload_handler(uint32_t *a, uint32_t *b, uint32_t length) {

  uint32_t i;

  for(i=0; i<length/sizeof(uint32_t); ++i) {   //xor msg and buffer
    a[i] = a[i] ^ b[i];
  }
  
  return 0;
}

int dsblock_payload_handler_omp(uint32_t *a, uint32_t *b, uint32_t length) {

  uint32_t i;
 #pragma omp parallel for schedule(static)
  for(i=0; i<length/sizeof(uint32_t); ++i) {   //xor msg and buffer
    a[i] = a[i] ^ b[i];
  }
  
  return 0;
}

int dsparity_payload_handler(uint32_t *a, uint32_t *b, uint32_t length)  {

  uint32_t i;
  const uint32_t mult = 36; 
  for(i=0; i<length/sizeof(uint32_t); ++i) {   //xor msg and buffer
    a[i] = (a[i] ^ b[i])*mult;
  }

  return 0;
}


int dsparity_payload_handler_omp(uint32_t *a, uint32_t *b, uint32_t length)  {

  uint32_t i;
  const uint32_t mult = 36; 
 #pragma omp parallel for schedule(static)
  for(i=0; i<length/sizeof(uint32_t); ++i) {   //xor msg and buffer
    a[i] = (a[i] ^ b[i])*mult;
  }

  return 0;
}

int accumulate_payload_handler(float* a, float* b, uint32_t length) {
     
    uint32_t i;
    for(i=0; i<length/sizeof(float); i+=2)
    {	
		a[i] = a[i]*b[i] - a[i+1]*b[i+1];
		a[i+1] = a[i]*b[i+1] - a[i+1]*b[i];
    }
    return 0;
}

int accumulate_payload_handler_omp(float* a, float* b, uint32_t length) {
     
    uint32_t i;
    #pragma omp parallel for schedule(static)
    for(i=0; i<length/sizeof(float); i+=2)
    {	
		a[i] = a[i]*b[i] - a[i+1]*b[i+1];
		a[i+1] = a[i]*b[i+1] - a[i+1]*b[i];
    }
    return 0;
}

int datatypes_payload_handler(void* a, void* b,  uint32_t length, uint32_t vlen, uint32_t stride) {

  for(uint32_t i=0; i<length/vlen; i++) 
    memcpy(b+(stride+vlen)*i,a+vlen*i,vlen);
  return 0;
}

int datatypes_payload_handler_omp(void* a, void* b,  uint32_t length, uint32_t vlen, uint32_t stride) {
 #pragma omp parallel for schedule(static)
  for(uint32_t i=0; i<length/vlen; i++) 
    memcpy(b+(stride+vlen)*i,a+vlen*i,vlen);
  return 0;
}

uint32_t main(int argc, char* argv[])
{
  if(argc<3)
	return 1;
  uint32_t N = 1000000,i,det=1;
  uint32_t handler = atoi(argv[1]);
  uint32_t data_length = atoi(argv[2]);
  uint32_t time = 0;
 
  uint32_t coef = data_length/32768;
  for(i=0;i<coef && N >=100000; i++){
    N/=10; det*=10;
  }

 struct timeval start, end;


  switch(handler)
  {
	case 0:  
	{
		uint32_t *a = (uint32_t*)malloc(data_length);
		uint32_t *b = (uint32_t*)malloc(data_length);
		fill_int(a,data_length/sizeof(uint32_t));		
		fill_int(b,data_length/sizeof(uint32_t));
		gettimeofday(&start, NULL);
			for(i=0; i<N; i++)
				dsblock_payload_handler_omp(a,b,data_length);
		gettimeofday(&end, NULL);
		uint32_t ps1 = ((end.tv_sec - start.tv_sec)*1e6 + (end.tv_usec - start.tv_usec))*det;
  	      	uint32_t ps2 = (uint32_t)-1;
		if(data_length<70000)
		{
			gettimeofday(&start, NULL);
			for(i=0; i<N; i++)
				dsblock_payload_handler(a,b,data_length);
			gettimeofday(&end, NULL);
			ps2 = ((end.tv_sec - start.tv_sec)*1e6 + (end.tv_usec - start.tv_usec))*det;
		}
		
		printf("%u\n",ps1<ps2 ? ps1 : ps2);
		free(a);
       		free(b);
		break;	
	}
	case 1: 
	{
		uint32_t *a = (uint32_t*)malloc(data_length);
		uint32_t *b = (uint32_t*)malloc(data_length);
		fill_int(a,data_length/sizeof(uint32_t));		
		fill_int(b,data_length/sizeof(uint32_t));
				gettimeofday(&start, NULL);
			for(i=0; i<N; i++)
				dsparity_payload_handler_omp(a,b,data_length);
		gettimeofday(&end, NULL);
		uint32_t ps1 = ((end.tv_sec - start.tv_sec)*1e6 + (end.tv_usec - start.tv_usec))*det;
  	      	uint32_t ps2 = (uint32_t)-1;
		if(data_length<70000)
		{
			gettimeofday(&start, NULL);
			for(i=0; i<N; i++)
				dsparity_payload_handler(a,b,data_length);
			gettimeofday(&end, NULL);
			ps2 = ((end.tv_sec - start.tv_sec)*1e6 + (end.tv_usec - start.tv_usec))*det;
		}
		
		printf("%u\n",ps1<ps2 ? ps1 : ps2);
		free(a);
       	        free(b);
		break;		
	}
	case 2:
	{
		float *a = (float *)malloc(data_length);
		float *b = (float *)malloc(data_length);
		fill_fp(a,data_length/sizeof(float));		
		fill_fp(b,data_length/sizeof(float));
		gettimeofday(&start, NULL);
			for(i=0; i<N; i++)
				accumulate_payload_handler_omp(a,b,data_length);
		gettimeofday(&end, NULL);
		uint32_t ps1 = ((end.tv_sec - start.tv_sec)*1e6 + (end.tv_usec - start.tv_usec))*det;
  	      	uint32_t ps2 = (uint32_t)-1;
		if(data_length<70000)
		{
			gettimeofday(&start, NULL);
			for(i=0; i<N; i++)
				accumulate_payload_handler(a,b,data_length);
			gettimeofday(&end, NULL);
			ps2 = ((end.tv_sec - start.tv_sec)*1e6 + (end.tv_usec - start.tv_usec))*det;
		}
		
		printf("%u\n",ps1<ps2 ? ps1 : ps2);
		
		free(a);
        	free(b);
		break;	
	}
	case 3:
	{
 		
	        uint32_t vlen = atoi(argv[3]);
		uint32_t stride = atoi(argv[4]);
		uint32_t *a = (uint32_t*)malloc(data_length+data_length);
		if(a==NULL)
			printf("Error");
		uint32_t *b = (uint32_t*)malloc(data_length+(stride+vlen)*data_length/(vlen-1)+data_length);
		if(b==NULL)
			printf("Error");
	
		gettimeofday(&start, NULL);
			for(i=0; i<N; i++)
				datatypes_payload_handler_omp(a,b,data_length,vlen,stride);
		gettimeofday(&end, NULL);
		uint32_t ps1 = ((end.tv_sec - start.tv_sec)*1e6 + (end.tv_usec - start.tv_usec))*det;
  	      	uint32_t ps2 = (uint32_t)-1;
		if(data_length<70000)
		{
			gettimeofday(&start, NULL);
			for(i=0; i<N; i++)
				datatypes_payload_handler(a,b,data_length,vlen,stride);
			gettimeofday(&end, NULL);
			ps2 = ((end.tv_sec - start.tv_sec)*1e6 + (end.tv_usec - start.tv_usec))*det;
		}
		
		printf("%u\n",ps1<ps2 ? ps1 : ps2);
		free(a);
       	        free(b);
		break;		
	}		
	default: return (uint32_t)-1;

  }

  return 0;
}
