#include "mem/cache/dummy_cache.hh"


DummyCache::DummyCache(DummyCacheParams *p)
    : MemObject(p), 
      cpuSidePort(p->name + ".cpu_side", this),
      memSidePort(p->name + ".mem_side", this)
{
}


BaseMasterPort& DummyCache::getMasterPort(const std::string& if_name, PortID idx){
    if (if_name == "mem_side") {
        return memSidePort;
    } else {
        return MemObject::getMasterPort(if_name, idx);
    }
}

BaseSlavePort& DummyCache::getSlavePort(const std::string& if_name, PortID idx){
    if (if_name == "cpu_side") {
        return cpuSidePort;
    } else {
        return MemObject::getSlavePort(if_name, idx);
    }
}



/* FROM MEMORY */
void DummyCache::recvTimingResp(PacketPtr pkt){

    //printf("[DummyCache] recvTimingResp from memory; pkt: %lu\n", pkt->getAddr());
    bool M5_VAR_USED success = cpuSidePort.sendTimingResp(pkt);

    assert(success);

}

void DummyCache::recvReqRetry(){
    assert(!wqueue.empty());

    while (!wqueue.empty()){
        PacketPtr waitingPkt = wqueue.front();

        //printf("[DummyCache] recvReqRetry from memory; pkt: %lu\n", waitingPkt->getAddr());
    
        bool M5_VAR_USED success = memSidePort.sendTimingReq(waitingPkt);       

        if (success) wqueue.pop();   
        else return;
    }
}

void DummyCache::recvRangeChange(){
    cpuSidePort.sendRangeChange();
}


/* FROM CPU */
void DummyCache::recvFunctional(PacketPtr pkt){
    memSidePort.sendFunctional(pkt);
}

bool DummyCache::recvTimingReq(PacketPtr pkt){

    assert(pkt->isRequest());

    //printf("[DummyCache] recvTimingReq from CPU; pkt: %lu; queued: %lu\n", pkt->getAddr(), wqueue.size());

    if (!wqueue.empty()){
        wqueue.push(pkt);    
    }else{
        bool M5_VAR_USED success = memSidePort.sendTimingReq(pkt);       
        if (!success) { wqueue.push(pkt); } 
        //printf("  result: %i\n", (int) success);
    }


    return true;
}

AddrRangeList DummyCache::getAddrRanges(){
    return memSidePort.getAddrRanges();
}


///////////////
//
// CpuSidePort
//
///////////////

AddrRangeList DummyCache::CpuSidePort::getAddrRanges() const{
    return cache->getAddrRanges();
}

bool DummyCache::CpuSidePort::recvTimingReq(PacketPtr pkt){
    return cache->recvTimingReq(pkt);
}

void DummyCache::CpuSidePort::recvFunctional(PacketPtr pkt){
    return cache->recvFunctional(pkt);
}



void DummyCache::CpuSidePort::recvRespRetry(){
    panic("rectRespRetry not expected!");
}

DummyCache* DummyCacheParams::create(){
    return new DummyCache(this);
}


///////////////
//
// MemSidePort
//
///////////////


bool DummyCache::MemSidePort::recvTimingResp(PacketPtr pkt){
    cache->recvTimingResp(pkt);
    return true;
}

void DummyCache::MemSidePort::recvReqRetry(){
    cache->recvReqRetry();
}

void DummyCache::MemSidePort::recvRangeChange(){
    cache->recvRangeChange();
}


