
#ifndef __DUMMY_CACHE__
#define __DUMMY_CACHE__


#include <vector>
#include <queue>
#include "mem/mem_object.hh"

#include "base/misc.hh" // fatal, panic, and warn
#include "params/DummyCache.hh"



class DummyCache : public MemObject{
  private:

    class CpuSidePort : public SlavePort{
      private:
        DummyCache *cache;

      public:
        CpuSidePort(const std::string& name, DummyCache *cache) :
            SlavePort(name, cache), cache(cache)
        { }

        AddrRangeList getAddrRanges() const override;

      protected:
        Tick recvAtomic(PacketPtr pkt) override{ 
            panic("recvAtomic unimpl."); 
        }

        void recvFunctional(PacketPtr pkt) override;

        bool recvTimingReq(PacketPtr pkt) override;

        void recvRespRetry() override;
    };

    class MemSidePort : public MasterPort{
      private:
        DummyCache * cache;

      public:
        MemSidePort(const std::string& name, DummyCache *cache) :
            MasterPort(name, cache), cache(cache)
        { }

      protected:
        bool recvTimingResp(PacketPtr pkt) override;

        void recvReqRetry() override;

        void recvRangeChange() override;
    };

    AddrRangeList getAddrRanges() const;

    void sendRangeChange();

    CpuSidePort cpuSidePort;

    MemSidePort memSidePort;

    std::queue<PacketPtr> wqueue;

  public:

    DummyCache(DummyCacheParams *params);

    void recvTimingResp(PacketPtr pkt);
    void recvReqRetry();
    void recvRangeChange();

    bool recvTimingReq(PacketPtr pkt);
    void recvFunctional(PacketPtr pkt);
    AddrRangeList getAddrRanges();


    BaseMasterPort& getMasterPort(const std::string& if_name, PortID idx = InvalidPortID) override;

    BaseSlavePort& getSlavePort(const std::string& if_name, PortID idx = InvalidPortID) override;
};

#endif /* __DUMMY_CACHE__ */
